#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <krb5.h>

typedef struct _keytab_entry_t {
	struct _keytab_entry_t *next;
	char *principal;
	unsigned long type;
	unsigned long timestamp;
	unsigned char kvno;
	unsigned short enctype;
	unsigned short keylen;
	unsigned char *key;
} keytab_entry_t;

keytab_entry_t *keytab_entries=NULL;


void 
read_keytab_file_502(FILE *fh)
{
	unsigned long rec_len;
	unsigned short prc;
	char pn[256],*pnp;
	keytab_entry_t *kte;


	while(1){
		/* record length */
		if(fread(&rec_len, 1, 4, fh)!=4){
			return;
		}
		rec_len=ntohl(rec_len);


		kte=malloc(sizeof(keytab_entry_t));
		kte->next=keytab_entries;
		keytab_entries=kte;

		/* number of components in principal */
		fread(&prc, 1, 2, fh);
		prc=ntohs(prc)+1; /* one component more than is stored */

		/* read all principal component into a string */
		pnp=pn;
		*pnp=0;
		while(prc--){
			unsigned short pclen;

			fread(&pclen, 1, 2, fh);
			pclen=ntohs(pclen);

			/* this is not the first component, insert a '/' */
			if(pnp!=pn){
				*pnp='/';
				pnp++;
			}
	
			fread(pnp, 1, pclen, fh);
			pnp+=pclen;
			*pnp=0;
		}
printf("PN:%s  ",pn);
		kte->principal=strdup(pn);

		/* principal type */
		fread(&kte->type, 1, 4, fh);
		kte->type=ntohl(kte->type);
printf("ptype:%x:%d  ",kte->type,kte->type);

		/* timestamp */
		fread(&kte->timestamp, 1, 4, fh);
		kte->timestamp=ntohl(kte->timestamp);

		/* kvno */
		fread(&kte->kvno, 1, 1, fh);
printf("kvno:%x:%d  ",kte->kvno,kte->kvno);

		/* enctype */
		fread(&kte->enctype, 1, 2, fh);
		kte->enctype=ntohs(kte->enctype);
printf("enctype:%x:%d  ",kte->enctype,kte->enctype);

		/* keylen */
		fread(&kte->keylen, 1, 2, fh);
		kte->keylen=ntohs(kte->keylen);
printf("keylen:%x:%d\n",kte->keylen,kte->keylen);

		/* key */
		kte->key=malloc(kte->keylen);
		fread(kte->key, 1, kte->keylen, fh);
	}

}

void 
read_keytab_file(char *filename)
{
	FILE *fh;
	unsigned short kt_vno;
	
	if(!(fh=fopen(filename, "r"))){
		fprintf(stderr, "Could not open keytab file :%s\n", filename);
		exit(10);
	}

	/* read keytab file version */
	fread(&kt_vno, 1, 2, fh);
	kt_vno=ntohs(kt_vno);
printf("kt version:%x\n",kt_vno);

	switch(kt_vno){
	case 0x0502:
		read_keytab_file_502(fh);
		break;
	}

	fclose(fh);
}

int enclen;
char encdata[2560];
char decdata[2560];
char all_zero[]={0,0,0,0,0,0,0,0};
void 
read_enc_data(char *filename)
{
	FILE *fh;
	unsigned short kt_vno;
	
	if(!(fh=fopen(filename, "r"))){
		fprintf(stderr, "Could not open encdata :%s\n", filename);
		exit(10);
	}
	enclen=fread(encdata,1,2550,fh);
printf("read encrypted data %d bytes\n",enclen);
	fclose(fh);
}
void print_data(unsigned char *str)
{
	int i;
	printf("DATA:");
	for(i=0;i<32;i++){
		printf(" %02x",str[i]);
	}
	printf("\n");
}
void enum_keytab_entries(char *principal, unsigned short enctype)
{
	keytab_entry_t *kte=keytab_entries;
	while(kte){
		if( (!strcmp(kte->principal, principal)) && (kte->enctype==enctype) ){
			krb5_keyblock kb;
			krb5_enc_data ki;
			krb5_data ko;
			int res;

			kb.magic=0;
			kb.enctype=kte->enctype;
			kb.length=kte->keylen;
			kb.contents=kte->key;

			ki.magic=0;
			ki.enctype=kte->enctype;
			ki.kvno=kte->kvno;
			ki.ciphertext.magic=0;
			ki.ciphertext.length=enclen;
			ki.ciphertext.data=encdata;

			ko.magic=0;
			ko.length=enclen;
			ko.data=decdata;

printf("Trying %s enc:%d kvno:%d\n",kte->principal,kte->enctype,kte->kvno);
			res=krb5_c_decrypt(NULL, &kb, 11, NULL, &ki, &ko);
			/*XXX 11 is AS_REP (frame 2)*/
printf("RES:%d\n",res);
print_data(decdata);
		}
		kte=kte->next;
	}
}

int main(int argc, char *argv[])
{

	read_keytab_file("krb5.keytab");
	read_enc_data("encpart.dat");
	print_data(encdata);
	enum_keytab_entries("CORE.UML/ronnie", 16);

}
